Library UniKey30;

Uses WinTypes, WinProcs, Strings,
     {$ifdef ver10} PxEngine, {$endif}
     {$ifdef ver15} PxEngine, {$endif}
     {$ifdef ver70} PxEngWin, PXMsg, {$endif}
     WinDos;

(*

    UNIKEY - unique record key generator
             Version 1.06: 93.02.01 - QUIKODE key generator
             Version 1.05: 92.11.20 - Better dialog box for lock collision
             Version 92.11.02 - Upgrade to Paradox Engine 3.0
             Version 92.04.09 - DelFile for deleting an ASCII file before updating
             Version 92.03.19 - EndSession call to release memory
             Version 92.03.06 - now includes Novell User Name! (environment)


    written by :  Jay Baker             CIS ID 73530,3146
                  6605 16th St N
                  Arlington, VA  22205

    Anyone may use this code, but if you have some nifty improvements,
    please send me a copy, too.

  This is a DLL for providing unique identification codes to any application
that needs a primary key for new records that is arbitrary.  The same DLL
can be used to maintain more than one sequence of unique keys.  Any number
of simultaneous users can obtain keys on a network because the Paradox
Engine is used to momentarily lock the master table of keys while a new one
is obtained.

  The basic technique used is to store the latest keys in a Paradox table.
This table has only one record, but may have more than one field.  Each
field in the table is for a different series of keys.  When the library
function is called to get a new key, the following sequence of steps occurs:

1.  The master key table is opened.
2.  The master key table is locked.  If the lock fails because another user
    has the table locked, then several tries are made (a message is also
    sent to the user to help randomize the delay between tries).
3.  A record buffer is opened, and the record is read.
4. *The particular key of interest is read from the record.
5. *The key is incremented.
6. *The new key is put back into the record.
7.  The record is written.
8.  The record buffer is closed.
9.  The table is unlocked.
10. The table is closed.
11. The new key is returned to the application as the function result.

  The items indicated by * are the only ones that must be overridden to
customize the key to any particular application.  Create a descendent of the
TUnique object and provide the GetKey, IncKey, and PutKey methods.

  This DLL contains two examples of creating new keys. The first uses a
4-byte alphanumeric string, and it is incremented alphanumerically, e.g.
V3GZ would become V3H0.  The second example is a more conventional 4-byte
integer, so 122473 would become 122474.

  Create the Paradox table as follows:

      Company ID      A4
      Serial Number   N

and then edit the table, putting the initial key values into each field.

  For ObjectVision applications, this DLL can be registered as follows:

    @REGISTER("@NEWCOMPANYID","CCH","","UniKey30.dll","NewA4",  0)
    @REGISTER("@NEWSERNO",    "JCH","","UniKey30.dll","NewLong",0)
    @REGISTER("@EndSession",  "I",  "","UniKey30.dll","EndSession",0)


Then call the function as @NEWSERNO("c:\dbdir\newkeys",2) whenever a new key
is needed.  This says use the second field in the NEWKEYS table to create a
new key according to the incrementing logic in the NewLong function.

  As many fields as desired may be put into the single Paradox table.  Each
field will control a separate series of unique keys.  Because the field
number is passed in each time, the same incrementing scheme can be used on
any number of different keys.  In fact, probably 99% of the unique keys in
the world will just use @NEWSERNO (or whatever you want to call it).

  UNIKEY assumes PxEngine is already initialized.  Otherwise you must
first call PXNetInit.  Alternatively, if you are using ObjectVision, just
put any link into your OV application, and the Engine will be initialized
automatically by OV.


3/5/92 - this unit now has a way to access the Novell network user name.
The user name is important for using as a key in OV.  This allows such
things as personalizing Rolodex records in a data base, by making the
user name part of the key, and setting a restricted range on the link.
The problem is that the Paradox Engine call to PXNetUserName only returns
the name that is saved in the PxEngCfg program, which is likely to be a
constant for all network users!

  During login on the network, a "login script" is run.  This is named
SYSLOGIN.COD.  The administrator can do things here that mere mortals can
not do.  One is to transfer special information to the dos environment.
For example,

  DOS SET USERNAM="%LOGIN_NAME"

is a reference to the hidden bindery file for a user.  This would put the
login name in the environment, like USERNAM=JBAKER.  Also available is
%FULL_NAME.

  This DLL depends on this special environment variable USERNAM which is
in no way standard.  It is peculiar to only one network that I know of.  The
other drawback in general is that it is not secure.  Anyone who knows how
can change an environment var at any time. It is not super secure, but it
is a start until OV provides the right user id.

  This DLL can be registered as follows:

    @REGISTER("@userid","C","","UniKey30.dll","NetUserName",  0)

  Then refer to @USERID anywhere in your OV trees to obtain the network
user id.


4/9/92 - DelFile is a procedure to delete any DOS file.  It is especially
useful to avoid the message when opening an existing ASCII file for writing
from an OV ASCII link.

  This DLL can be registered as follows:

    @REGISTER("@erase","JC","","UniKey30.dll","DelFile",0)

  Then refer to @ERASE anywhere in your OV trees to delete a DOS file.


@EndSession:
  performs any housekeeping required before ending session.

  To register this function, use

  @ASSIGN(path,"f:\xyz\")
  @REGISTER("@EndSession","I","",path&"UniKey30.dll","EndSession",0)


11/20/92: More experience on network lock collisions has led to use
of Retry and Cancel buttons on the lock dialog box.  Also, it was learned
that an open table that has failed to be locked must be closed before
giving user the dialog box - otherwise no one else can obtain the FL lock
either.

02/01/93: QUIKODE is a very fast yet simple indexing scheme for names.
Using a few letters of the first and last names, it can usually find an
individual in a very large list quite quickly.  Names with high collisions
in the coding space, or names with common nicknames that are very different
from the original (ROBERT/BOB), are given unique codes.  The result is a
5-byte code that can be indexed much more effectively than the whole last
name and first name.  QUIKODE is even tolerant of some misspellings.

  This DLL can be registered as follows:

    @REGISTER("@Quikode","CCC","","UniKey30.dll","QuiKode",0)

  Then refer to @quikode(firstname,lastname) anywhere in your OV trees to
  create the 5-byte code for the name index.


*)

const
  ver='1.06';
  ErrCaption='UniKey '+ver;

type

  {This is the root object for creating unique keys}

  PUnique=^TUnique;
  TUnique=object
    constructor init(FileName:pchar;FieldNo:word);
    destructor done;                                       virtual;
    procedure abstract;
    function  CheckError(Code: Integer): Boolean;
    function  GetKey(RecHandle:RecordHandle):boolean;      virtual;
    procedure IncKey;                                      virtual;
    procedure MakeNewKey;
    function  OpenLockWait(var hTbl:TableHandle):integer;  virtual;
    procedure PutKey(RecHandle:RecordHandle);              virtual;
    function  SameSpecs(FileName:pchar;FieldNo:word):boolean;

    private
    ufilename:array[0..79] of char; {name of Paradox table containing id}
    ufieldno:integer;               {field number in Paradox table}
    function  WaitMsg(err:integer):integer;
    end;

  {Following are two examples of descendent objects for particular kinds
  of keys}

  tstr5=array[0..4] of char;

  PNewA4=^TNewA4;
  TNewA4=object(TUnique)
    x:tstr5; {a 4-byte alphanumeric code that will be incremented each call}

    function  GetKey(RecHandle:RecordHandle):boolean;  virtual;
    procedure IncKey;                                  virtual;
    procedure PutKey(RecHandle:RecordHandle);          virtual;
    end;

  PNewLong=^TNewLong;
  TNewLong=object(TUnique)
    n:longint; {numeric serial number that will be incremented each call}

    function  GetKey(RecHandle:RecordHandle):boolean;  virtual;
    procedure IncKey;                                  virtual;
    procedure PutKey(RecHandle:RecordHandle);          virtual;
    end;

function int2str(i:longint):string;
var
  s:string;
begin
  str(i,s);
  int2str:=s
  end; {int2str}

{---------------------------- TUnique -------------------------------------}

constructor TUnique.init(FileName:pchar;FieldNo:word);
begin
  StrCopy(ufilename,Filename);
  ufieldno:=fieldno;
  end; {init}

destructor TUnique.done;
begin
  end; {done}

procedure TUnique.abstract;
begin
  MessageBox(GetFocus, 'Abstract method called - programming error', ufilename, mb_OK);
  end; {abstract}

function TUnique.CheckError(Code: Integer): Boolean;
var
  a:array[0..99] of char;
begin
  if Code <> PXSuccess then begin
    StrPCopy(a,'File '+StrPas(ufilename)+': '+StrPas(PXErrMsg(Code)));
    MessageBox(GetFocus, a, ErrCaption, mb_OK);
    end;
  CheckError := Code = PXSuccess;
  end; {checkerror}

function TUnique.GetKey(RecHandle:RecordHandle):boolean;
begin
  abstract
  end; {getkey}

procedure TUnique.IncKey;
begin
  abstract
  end; {IncKey}

function TUnique.WaitMsg(err:integer):integer;
var
  ans:integer;
  a:array[0..300] of char;
begin
  ans:=IDRETRY;
  if err<>PXSuccess then begin
    StrPCopy(a,StrPas(ufilename)+': '+
               'Error '+int2str(err)+': '+StrPas(PXErrMsg(err))+'.'#13#13+
               'Table may be temporarily locked by another user.'#13+
               'Press Retry to try again, '+
               'or Cancel to terminate request for new ID.');
    ans:=MessageBox(GetFocus,a,ErrCaption, mb_RetryCancel);
    end;
  waitmsg:=ans
  end; {waitmsg}

function TUnique.OpenLockWait(
  var hTbl:TableHandle)
	       :integer;
{tries to Open a table, waits a little while for its availability}
var
  err:integer;
  i:integer;
  openok:boolean;
begin
  repeat {this outer loop is controlled by user if there is an error}
    i:=0;
    repeat {this inner loop is automatically tried 5 times by computer}
      err:=PXTblOpen(ufilename, hTbl, 0, True);
      openok:=err=PXSuccess; {do not retry if table can not even be opened}
      if openok then begin
        err:=PXNetTblLock(hTbl,FL);
        if err<>PXSuccess then
          PXTblClose(hTbl);
        end;
      inc(i)
    until (err=PXSuccess) or not openok or (i>5);
  until (err=PXSuccess) or (WaitMsg(err)=IDCANCEL);
  OpenLockWait:=err
  end; {Openlockwait}

procedure TUnique.MakeNewKey;

  procedure mnk2(hTbl:TableHandle);
  var
    RecHandle: RecordHandle;
  begin
    if CheckError(PXRecLast    (hTbl           )) and
       CheckError(PXRecBufOpen (hTbl, RecHandle)) then begin
	 if CheckError(PXRecGet(hTbl, RecHandle)) and
	    GetKey(RecHandle) then begin
	     IncKey;
	     PutKey(RecHandle);
	     PXRecUpdate(hTbl, RecHandle);
	     end;
	 PXRecBufClose(RecHandle);
	 end;
    end; {mnk2}

var {makenewkey}
  hTbl: TableHandle;

begin {makenewkey}
  if   OpenLockWait(hTbl)=PXSuccess then begin
    mnk2(hTbl);
    CheckError(PXNetTblUnlock(hTbl,FL));
    PXTblClose(hTbl);
    end;
  end; {MakeNewKey}

procedure TUnique.PutKey(RecHandle:RecordHandle);
begin
  abstract
  end; {Putkey}

function TUnique.SameSpecs(FileName:pchar;FieldNo:word):boolean;
begin
  SameSpecs:=(StrComp(FileName,ufilename)=0) and (FieldNo=ufieldno)
  end; {samespecs}

{---------------------------- TNewA4 -----------------------------------}

function TNewA4.GetKey(RecHandle:RecordHandle):boolean;
begin
  getkey:=CheckError(PXGetAlpha(RecHandle, ufieldno, SizeOf(x), x))
  end; {getkey}

procedure TNewA4.IncKey;
var
  i:integer;
  carry:boolean;
  c:tstr5;
begin
  carry:=true;
  StrCopy(c,x);
  for i:=3 downto 0 do if carry then
    case c[i] of
      '0'..'8',
      'A'..'Y':begin
               repeat
                 c[i]:=succ(c[i]); carry:=false
               until c[i]<>'O'; {avoid confusion with zero}
               end;
      '9'     :begin c[i]:='A';        carry:=false end;
      'Z'     :begin c[i]:='0';        carry:=true  end;
      else c[i]:='0'; {if any field is garbage}
      end;
  c[4]:=#0;
  StrCopy(x,c);
  end; {IncKey}

procedure TNewA4.PutKey(RecHandle:RecordHandle);
begin
  CheckError(PXPutAlpha(RecHandle, ufieldno, x));
  end; {Putkey}

{---------------------------- TNewLong -----------------------------------}

function TNewLong.GetKey(RecHandle:RecordHandle):boolean;
begin
  getkey:=CheckError(PXGetLong(RecHandle, ufieldno, n))
  end; {getkey}

procedure TNewLong.IncKey;
begin
  inc(n)
  end; {IncKey}

procedure TNewLong.PutKey(RecHandle:RecordHandle);
begin
  CheckError(PXPutLong(RecHandle, ufieldno, n));
  end; {Putkey}

{----------------------------------------------------------------------}

const
  CompanyID:PNewA4 = nil; {use typed constant because address is returned}

function NewA4(FileName:pchar;FieldNo:word):pchar; export;
begin
  if (CompanyID<>nil) and not CompanyID^.SameSpecs(FileName,FieldNo) then begin
    dispose(CompanyID,done); CompanyID:=nil end;
  if CompanyID=nil then
    CompanyID:=new(PNewA4,init(FileName,FieldNo));
  CompanyID^.x[0]:=#0;
  CompanyID^.MakeNewKey;
  NewA4:=@CompanyID^.x; {if null is returned, then something went wrong}
  end; {NewA4}

function NewLong(FileName:pchar;FieldNo:word):longint; export;
var
  SerNo:PNewLong;
begin
  SerNo:=new(PNewLong,init(FileName,FieldNo));
  SerNo^.n:=0;
  SerNo^.MakeNewKey;
  NewLong:=SerNo^.n; {if zero is returned, then something went wrong}
  dispose(SerNo,done);
  end; {NewLong}

function DelFile(FileName:pchar):longint; export;
var
  F:  file;
begin
  Assign(F, StrPas(filename));
  {$I-} Reset(F); {$I+}
  if IOResult=0 then begin
    Close(F);
    Erase(F);
    end;
  end; {delfile}

function NetUserName:PChar; export;
const
  username:array[0..20] of char='';
var
  code:integer;
  p:PChar;
begin
  p:=WinDos.GetEnvVar('USERNAM');
  if p=nil then
    StrCopy(username,'?user?')
  else
    StrLCopy(username,p,sizeof(username)-1);

(*
------ this one should work, but doesn't - OV doesn't initialize the
       Engine right, IMHO.  It forces the user name to what is in Win.Ini

  code:=PXNetUserName(sizeof(username)-1,username);
  if code=PXSuccess then
    username[StrLen(username)]:=#0
  else begin
    username[0]:=#0;
    MessageBox(GetFocus, PXErrMsg(code), 'Net User Name', mb_OK);
    end;
*)

  NetUserName:=@username;
  end; {NetUserName}

function EndSession:integer; export;
{Summary
  perform any housekeeping required before ending session.

  To register this function, use

  @ASSIGN(path,"f:\xyz\")
  @REGISTER("@EndSession","I","",path&"UniKey30.dll","EndSession",0)

  }
begin
  if CompanyID<>nil then begin
    dispose(CompanyID,done); CompanyID:=nil end;
  end; {endsession}

function upstr(a:string):string;
var
  i:integer;
begin
  for i:=1 to length(a) do
    a[i]:=upcase(a[i]);
  upstr:=a
  end; {upstr}

function trim(x:string):string;
begin
  while x[length(x)]=' ' do dec(byte(x[0]));
  trim:=x
  end; {trim}

function QuiKode(first,last:PChar):PChar; export;
{Summary
  Change first and last names into a 5-byte code for easy indexing}
const
  kode:array[0..4] of char='';
var
  fnm,lnm:string[40];
  n1:string[3];
  n2:string[2];

begin
  fnm:=StrPas(first);
  lnm:=StrPas(last);
  fnm:=upstr(trim(fnm));
  lnm:=upstr(trim(lnm));

  if length(lnm)<3 then
    n1:=lnm+copy('   ',1,3-length(lnm))
  else
    n1:=lnm[1]+lnm[(length(lnm)+1) div 2]+lnm[length(lnm)];

  if fnm='' then
    n2:=''
  else
    if fnm='JOHN'   then n2:='X1' else

    if fnm='DAVID'  then n2:='X2' else
    if fnm='DAVE'   then n2:='X2' else

    if fnm='MICHAEL'then n2:='X3' else
    if fnm='MIKE'   then n2:='X3' else

    if fnm='ROBERT' then n2:='X4' else
    if fnm='BOB'    then n2:='X4' else

    if fnm='STEVE'  then n2:='X5' else
    if fnm='STEVEN' then n2:='X5' else
    if fnm='STEPHEN'then n2:='X5' else

    if fnm='RICHARD'then n2:='X6' else
    if fnm='DICK'   then n2:='X6' else

    if fnm='DONALD' then n2:='X7' else
    if fnm='DON'    then n2:='X7' else

    if fnm='CHARLES'then n2:='X8' else
    if fnm='CHARLEY'then n2:='X8' else
{   if fnm='CHARLIE'then n2:='X8' else}
    if fnm='CHUCK'  then n2:='X8' else

    if fnm='CHRIS'      then n2:='X9' else
    if fnm='CHRISTOPHER'then n2:='X9'

(* need lots of fixing to implement now
    if fnm='WILLIAM'    then n2:='XA' else
    if fnm='BILL'       then n2:='XA'
*)
    else
      n2:=fnm[1]+fnm[length(fnm)];


  if (fnm='') and (lnm='') then
    StrCopy(kode,'')
  else
    StrPCopy(kode,n1+n2);

  QuiKode:=@kode
  end; {quikode}

exports
  NewA4, NewLong,
  NetUserName, DelFile, EndSession,
  QuiKode;

begin
end.
